/*
 *  LSynth Service Tool
 *  Derived from LSynth 2.0 by Kevin Clague
 *  Jim DeVona 10 April 2007
 *
 *  The main feature of this LSynth variant is stdin/stdout "filter" capability.
 *  Additional changes include case preservation, whitespace insensitivity,
 *  experimental support for whitespace in constraint part filenames (silly),
 *  greater tolerance for non-synth lines within synth blocks, relaxed CRLF
 *  input parsing, a tad more error handling/acknowledgement, and sundry
 *  problems surely introduced by my wanton dismissal of miscellaneous code.
 *  I have also added many comments to assist my understanding of the program's flow.
 *
 *  This adaptation is based on the lsynth_source_2_0 release available at:
 *  <http://sourceforge.net/project/showfiles.php?group_id=71694&package_id=71310&release_id=154605>
 *  If I had realized that a new version of lsynth is actively under development,
 *  I might have made better use of my time by basing this adapation on the current
 *  code rather than an obsolete branch. I leave the prospect of integration to others.
 *
 *  These changes were made to facilitate the creation of an LSynth Service
 *  for Mac OS X (services are system-wide text-processing extensions that can
 *  be applied almost anywhere there is editable text). Applications that use
 *  proprietary text editing fields exhibit varied support for this service,
 *  but it works well with all standard text fields. The lsynthst utility is
 *  packaged as a service using Jesper's ThisService:
 *  <http://wafflesoftware.net/thisservice/>
 *
 *  Some line-ending chomping idioms:
 *
 *      int spanned;
 *      spanned = strcspn(line, LINE_ENDING_CHARS);
 *      line[spanned] = '\0';
 *      
 *      char *c;
 *      c = strpbrk(line, LINE_ENDING_CHARS);
 *      if (c != NULL) *c = '\0';
 *
 *  Even nicer is a custom fgets that chomps that cruft itself. Another
 *  advantage of a custom fgets is recognition of \r by itself as a line ending.
 *  I successfully tried such a function derived from recent Lugnet discussion,
 *  but opted not to use it (for now) because the only need I found for \r support
 *  was BBEdit/TextWrangler integration (these editors expect all service text
 *  transactions to done with \r line endings, frustrating evidence of their
 *  venerable past). Writing a separate wrapper script for these editors proved
 *  simpler than burdening lsynth with additional logic to output \r endings.
 *  Nevertheless, it seems that a purpose-built fgets would consolidate and
 *  extend lsynth's functionality nicely. Towards that end, some fgets ideas:
 *
 *      Travis Cobbs and Chris Phillips: <http://news.lugnet.com/cad/?n=14545>
 *      Microsoft: <http://research.microsoft.com/invisible/src/crt/fgets.c.htm>
 *      Darwin 6.8: <http://www.opensource.apple.com/darwinsource/10.2.8/Libc-262.3.2/stdio/fgets.c>
 *
 *  So, I hope someone finds this useful. I do.
 *
 */

#include "lsynth.h"
#include "tube.h"
#include "band.h"
#include <ctype.h>

/* list of hose type names (with trailing spaces) */
static char *tube_types[] = {
  ALL_HOSES,
  "",
};

/* list of band type names (with trailing spaces) */
static char *band_types[] = {
  ALL_BANDS,
  "",
};

/*
 *  strupper
 *
 *  Capitalize every character in s.
 *
 *  Parameters:
 *    s, terminated string to capitalize (modified in place)
 *
 *  Return:
 *    none
 *
 */
void strupper(char *s) {
    char *p;
    for (p = s; *p != '\0'; p++) {
        *p = toupper(*p);
    }
}

/*
 *  trimtrailing
 *
 *  Trim trailing whitespace from s.
 *
 *  Parameters:
 *    s, terminated string to trim (modified in place)
 *
 *  Return:
 *    none
 *
 */
void trimtrailing(char *s) {
    int index;    
    for (index = strlen(s) - 1; index >= 0; index--) {
        if (isblank(s[index])) {
            s[index] = '\0';
        } else {
            break;
        }
    }
}

/*
 *  Matches
 *
 *  Test whether a string matches a simple scan pattern. The pattern is given
 *  in upper case but the line is matched case insensitively. As matching is
 *  performed with sscanf, whitespace matches any span of whitespace, and
 *  leading and trailing whitespace additionally matches no whitespace.
 *
 *  Parameters:
 *    line, terminated input string to test
 *    pattern, terminated template string (capitalized; must end with %n)
 *
 *  Return:
 *    0 no match (or empty pattern)
 *    nonzero span of successful match
 *
 */
int Matches(const char *line, const char *pattern) {
    static char capline[MAX_LINE_LENGTH];
    int span = 0;

    /* compare a capitalized copy of the line */
    strcpy(capline, line);
    strupper(capline);
        
    /* try to scan the pattern from the line */
    sscanf(capline, pattern, &span);
    
    return span;
}

/*
 *  WriteLine
 *
 *  A shortcut to output a line with the proper line ending.
 *  Mainly for clarity and to help consolidate line ending shenanigans.
 *
 *  Parameters:
 *    line, terminated string to write (no EOL characters)
 *    dst, destination file
 *
 *  Return:
 *    none
 *
 */
void WriteLine(const char *line, FILE *dst) {
    fputs(line, dst);
    fputs(LINE_ENDING, dst);
}

/*
 *  skip_synthesized
 *
 *  Skip through the remainder of a previously synthesized section.
 *
 *  Parameters:
 *    dat, input file
 *    line, partial line containing start of synthesized section (may be changed in place)
 *    sizeof_line, max length of partial line (MAX_LINE_LENGTH less any starting offset)
 *
 *  Return:
 *    0 ok (terminated SYNTHESIZED section skipped)
 *   -1 error (unterminated SYNTHESIZED section)
 *
 */
int skip_synthesized(FILE *dat, char *line, int sizeof_line) {
    char *eol;
    
    /* skim through this previously-synthesized section */
    while (fgets(line, sizeof_line, dat)) {
        
        /* chomp line ending characters (not really necesssary) */
        eol = strpbrk(line, LINE_ENDING_CHARS);
        if (eol != NULL) *eol = '\0';
        
        /* return successfully once we hit the end of the previously-synthesized section */       
        if (Matches(line, " 0 SYNTH SYNTHESIZED END %n")) {
            return 0;
        }
    }
    
    return -1;
}
/*
 *  synth_tube_like
 *
 *  Synthesize a hose-like component.
 *
 *  Parameters:
 *    type, string hose type declaration
 *    resid, string color declaration
 *    dat, input file
 *    dst, output file
 *
 *  Return:
 *    0 ok
 *   -1 error (scan error, constraint limit exceeded, synthesis failed)
 *
 */
static int synth_tube_like(char *type, char *resid, FILE *dat, FILE *dst)
{
    char line[MAX_LINE_LENGTH];
    char *line_position;
    LSL_part_usage constraints[MAX_HOSE_CONSTRAINTS];
    int constraint_n = 0;
    char start_type[MAX_PART_PATH];
    float x,y,z, a,b,c, d,e,f, g,h,i;
    int hose_color;
    int temp_color;
    int rc = 0;
    int ghost = 0;
    int span_ghost;
    float length;
    char units[64];
    char *eol;
    int hide = 1;
    
    /* By default, constraint parts are hidden in output (hide = 1 above). */
    /* Consequentially, re-running LSynth on the unmodified output will */
    /* not see the constraints and therefore actually delete the synthesized parts. */
    /* Use 0 SYNTH SHOW to keep these parts visible to facilitate lsynth reruns. */
    
    /* read hose color */
    if (sscanf(resid, " %d ", &hose_color) != 1) {
        return -1;
    }
    
    /* initialize synth_tube arguments that we don't seem to change */
    length = 0;
    units[0] = '\0';
    
    /* we handle lines within synth hose blocks here instead of main */
    while (fgets(line, sizeof(line), dat)) {
        
        /* chomp line ending characters */
        eol = strpbrk(line, LINE_ENDING_CHARS);
        if (eol != NULL) *eol = '\0';
        if (line[0] == '\0') {
            continue;
        }
        
        /* honor MLCad "ghost" mode; use line_position instead of line after here */
        line_position = line;
        span_ghost = Matches(line, " 0 GHOST %n");
        if (span_ghost > 0) {
            ghost = 1;
            line_position += span_ghost;
        }
        
        /* read in a part, stuffing everything it can into the part name */
        if (sscanf(line_position, "1 %d %f %f %f %f %f %f %f %f %f %f %f %f %[^" LINE_ENDING_CHARS "]",
                &temp_color, &x,&y,&z, &a,&b,&c, &d,&e,&f, &g,&h,&i, start_type) == 14) {
            
            /* trim trailing whitespace from name (scanf handles leading) */
            trimtrailing(start_type);
            
            /* only special constraint parts are read */
            if (Matches(start_type, " LS00.DAT %n")) {
            
                /* output hidden parts with comment prefix */
                if (hide) {
                    fputs("0 ", dst);
                }
                
                /* output the part */
                WriteLine(line_position, dst);
                
                /* record characteristics */
                constraints[constraint_n].loc.x = x;
                constraints[constraint_n].loc.y = y;
                constraints[constraint_n].loc.z = z;
                constraints[constraint_n].orient[0][0] = a;
                constraints[constraint_n].orient[0][1] = b;
                constraints[constraint_n].orient[0][2] = c;
                constraints[constraint_n].orient[1][0] = d;
                constraints[constraint_n].orient[1][1] = e;
                constraints[constraint_n].orient[1][2] = f;
                constraints[constraint_n].orient[2][0] = g;
                constraints[constraint_n].orient[2][1] = h;
                constraints[constraint_n].orient[2][2] = i;
                strcpy(constraints[constraint_n].type, start_type);
                strupper(constraints[constraint_n].type); /* capitalize internally */
                constraint_n++;
                
            } else {
                
                /* just echo non-LS00.dat parts */
                WriteLine(line_position, dst);
            }
        } else if (Matches(line_position, " 0 SYNTH HIDE %n") || Matches(line_position, " 0 WRITE SYNTH HIDE %n")) {

            /* start hiding output */
            WriteLine(line_position, dst);
            hide = 1;
          
        } else if (Matches(line_position, " 0 SYNTH SHOW %n") || Matches(line_position, " 0 WRITE SYNTH SHOW %n")) {
                    
            /* stop hiding output */
            WriteLine(line_position, dst);
            hide = 0;
          
        } else if (Matches(line_position, " 0 SYNTH SYNTHESIZED BEGIN %n")) {
            
            /* skip previously-synthesized section */
            skip_synthesized(dat, line_position, strlen(line_position));
            
        } else if (Matches(line_position, " 0 SYNTH END %n") || Matches(line_position, " 0 WRITE SYNTH END %n")) {
            
            /* perform the actual synthesis before closing the synth section */
            rc = synth_tube(type, constraint_n, constraints, length, units, hose_color, dst, ghost);
            WriteLine(line_position, dst);
            
            /* exit this synth handler loop */
            break;
            
        } else {
        
            /* echo anything that's not a part or a relevant synth command */
            WriteLine(line_position, dst);
        }
        
        /* check if we have somehow exceeded the our allotment of constraints */
        if (constraint_n >= MAX_HOSE_CONSTRAINTS) {
            fprintf(stderr, "Maximum number of hose constraints exceeded: %d\n", constraint_n);
            return -1;
        }
    }
    
    return rc;
}

/*
 *  synth_band_like
 *
 *  Synthesize a band-like component.
 *
 *  Parameters:
 *    type, string band type declaration
 *    resid, string color declaration
 *    dat, input file
 *    dst, output file
 *
 *  Return:
 *    0 ok
 *   -1 error (scan error, constraint limit exceeded, synthesis failed)
 *
 */
static int synth_band_like(char *type, char *resid, FILE *dat, FILE *dst)
{
    char line[MAX_LINE_LENGTH];
    char *line_position;
    LSL_band_constraint constraints[MAX_BAND_CONSTRAINTS];
    int constraint_n = 0;
    int color = 0;
    int temp_color;
    char start_type[MAX_PART_PATH];
    float x,y,z, a,b,c, d,e,f, g,h,i;
    int hide = 0;    
    int ghost = 0;
    int span_ghost;
    int rc = 0;
    char *eol;
    
    /* read band color */
    if (sscanf(resid," %d ", &color) != 1) {
        return -1;
    }
        
    /* we handle lines within synth band blocks here instead of main */
    while (fgets(line, sizeof(line), dat)) {
        
        /* chomp line ending characters */
        eol = strpbrk(line, LINE_ENDING_CHARS);
        if (eol != NULL) *eol = '\0';
        if (line[0] == '\0') {
            continue;
        }
        
        /* honor MLCad "ghost" mode; use line_position instead of line after here */
        line_position = line;
        span_ghost = Matches(line, " 0 GHOST %n");
        if (span_ghost > 0) {
            ghost = 1;
            line_position += span_ghost;
        }
        
        /* read in a part, stuffing everything it can into the part name */
        if (sscanf(line_position, " 1 %d %f %f %f %f %f %f %f %f %f %f %f %f %[^" LINE_ENDING_CHARS "]",
                &temp_color, &x,&y,&z, &a,&b,&c, &d,&e,&f, &g,&h,&i, start_type) == 14) {
            
            /* trim trailing whitespace from name (scanf handles leading) */
            trimtrailing(start_type);
            
            /* output hidden parts with comment prefix */
            if (hide) {
                fputs("0 ", dst);
            }
                        
            /* output the part */
            WriteLine(line_position, dst);
            
            /* record characteristics */
            constraints[constraint_n].part.color = color; /* synth color instead of temp_color? */
            constraints[constraint_n].part.loc.x = x;
            constraints[constraint_n].part.loc.y = y;
            constraints[constraint_n].part.loc.z = z;            
            constraints[constraint_n].part.orient[0][0] = a;
            constraints[constraint_n].part.orient[0][1] = b;
            constraints[constraint_n].part.orient[0][2] = c;
            constraints[constraint_n].part.orient[1][0] = d;
            constraints[constraint_n].part.orient[1][1] = e;
            constraints[constraint_n].part.orient[1][2] = f;
            constraints[constraint_n].part.orient[2][0] = g;
            constraints[constraint_n].part.orient[2][1] = h;
            constraints[constraint_n].part.orient[2][2] = i;
            strcpy(constraints[constraint_n].part.type, start_type);
            strupper(constraints[constraint_n].part.type); /* capitalize internaly */
            constraint_n++;
          
        } else if (Matches(line_position, " 0 SYNTH INSIDE %n") || Matches(line_position, " 0 WRITE SYNTH INSIDE %n")) {
          
            /* "inside" constraint */
            WriteLine(line_position, dst);
            strcpy(constraints[constraint_n].part.type,"INSIDE");
            constraint_n++;
          
        } else if (Matches(line_position, " 0 SYNTH OUTSIDE %n") || Matches(line_position, " 0 WRITE SYNTH OUTSIDE %n")) {

            /* "outside" constraint */
            WriteLine(line_position, dst);
            strcpy(constraints[constraint_n].part.type,"OUTSIDE");
            constraint_n++;
          
        } else if (Matches(line_position, " 0 SYNTH CROSS %n") || Matches(line_position, " 0 WRITE SYNTH CROSS %n")) {
                    
            /* "cross" constraint */
            WriteLine(line_position, dst);
            strcpy(constraints[constraint_n].part.type,"CROSS");
            constraint_n++;
          
        } else if (Matches(line_position, " 0 SYNTH HIDE %n") || Matches(line_position, " 0 WRITE SYNTH HIDE %n")) {

            /* start hiding output */
            WriteLine(line_position, dst);
            hide = 1;
          
        } else if (Matches(line_position, " 0 SYNTH SHOW %n") || Matches(line_position, " 0 WRITE SYNTH SHOW %n")) {
                    
            /* stop hiding output */
            WriteLine(line_position, dst);
            hide = 0;
          
        } else if (Matches(line_position, " 0 SYNTH SYNTHESIZED BEGIN %n")) {

            /* skip previously-synthesized section */
            skip_synthesized(dat, line_position, strlen(line_position));
            
        } else if (Matches(line_position, " 0 SYNTH END %n") || Matches(line_position, " 0 WRITE SYNTH END %n")) {
           
            /* perform the actual synthesis before closing the synth section */
           rc = synth_band(type, constraint_n, constraints, color, dst, ghost);
           WriteLine(line_position, dst);
           
           /* exit this synth handler loop */
           break;
           
        } else {
        
            /* echo anything that's not a part or a relevant synth command */
            WriteLine(line_position, dst);
        }
        
        /* check if we have somehow exceeded the our allotment of constraints */
        if (constraint_n >= MAX_BAND_CONSTRAINTS) {
            fprintf(stderr, "Maximum number of band constraints exceeded: %d\n", constraint_n);
            return -1;
        }
    }
    
    /* return synth_band's result */
    return rc;
}

/*
 *  main
 *  
 *  LSynth main function. Reads arguments and sets up IO.
 *  Loops through input lines, echoing them to output, and invokes
 *  synthesizer functions which may insert additional lines where
 *  synth sections are encountered.
 *  
 *  Parameters:
 *    argc, number of arguments
 *    argv, list of arguments (progname [-v] [dat_name dst_name])
 *  
 *  Return:
 *    0 success
 *    1 argument error
 *   -1 file or synth error
 */
int main(int argc, char *argv[])
{
    char *dat_name = argv[1];
    char *dst_name = argv[2];
    FILE *dat;
    FILE *dst;
    char line[MAX_LINE_LENGTH];
    char *line_position;
    int span_sb, span_wsb;
    int type_index;
    int rc = 0;
    int spanned;
    char type[MAX_TYPE_LENGTH + 1];
    char *eol;
    
    /* invoked with no arguments, filter stdin to stdout */
    if (argc == 1) {
        dat = stdin;
        dst = stdout;
    }
    else {
        
        /* otherwise, proceed as usual */
        printf("LSynth Service Tool version 2.0 by Jim DeVona (experimental)\n");
        printf("Derived from LSynth version 2.0 by Kevin Clague, kevin_clague@yahoo.com\n");
        
        /* just interested in version? */
        if (argc == 2 && strcmp(argv[1], "-v") == 0) {
            return 1;
        }       
        
        /* check for standard arguments */
        if (argc != 3) {
            printf("Usage: %s [-v] [in_file out_file]\n", argv[0]);
            printf("       lsynthcp -v                Report version\n");
            printf("       lsynthcp in_file out_file  Synthesize in_file to out_file\n");
            printf("       lsynthcp                   Synthesize stdin to stdout\n");
            return 1;
        }
        
        /* open input file */
        dat = fopen(dat_name,"r");
        if (dat == NULL) {
            printf("%s: Failed to open file %s for reading\n",argv[0],dat_name);
            return -1;
        }
        
        /* open output file */
        dst = fopen(dst_name,"w");
        if (dst == NULL) {
            printf("%s: Failed to open file %s for writing\n",argv[0],dst_name);
            return -1;
        }
    }
    
    /* process each line of input */
    while (fgets(line, sizeof(line), dat)) {

        /* chomp line ending characters */
        eol = strpbrk(line, LINE_ENDING_CHARS);
        if (eol != NULL) *eol = '\0';
        if (line[0] == '\0') {
            continue;
        }
        
        /* mirror the line to output with standardized line endings */
        WriteLine(line, dst);
                
        /* Matches uses sscanf; %n returns the number of matching characters scanned */
        span_sb = Matches(line, " 0 SYNTH BEGIN %n");
        span_wsb = Matches(line, " 0 WRITE SYNTH BEGIN %n");
        
        /* if the scan made it to %n, the span will be nonzero */
        if (span_sb > 0 || span_wsb > 0) {
            rc = 0;
            
            /* spans can conveniently be used to get to start of type, accounting for variable whitespace */
            line_position = line;
            if (span_sb > 0) {
                line_position += span_sb;
            }
            if (span_wsb > 0) {
                line_position += span_wsb;
            }
                        
            /* case insensitive type comparison with span to expected color location */
            sscanf(line_position, " %s %n", type, &spanned);
            strcat(type, " "); /* since type definitions include trailing spaces */
            strupper(type);
            
            /* check to see if the remainder of the line is a known synth tube command */
            for (type_index = 0; tube_types[type_index][0] != '\0'; type_index++) {
                if (strcmp(type, tube_types[type_index]) == 0) {
                    rc = synth_tube_like(type, line_position + spanned, dat, dst);
                    break;
                }
            }
                        
            /* check to see if the remainder of the line is a known synth band command */
            for (type_index = 0; band_types[type_index][0] != '\0'; type_index++) {
                if (strcmp(type, band_types[type_index]) == 0) {
                    rc = synth_band_like(type, line_position + spanned, dat, dst);
                    break;
                }
            }
            
            /* if synth failed, should we abort or what? */
            if (rc != 0) {
                break;
            }
        }        
    }
    
    /* so, how'd it go? */
    if (dst != stdout) {
        if (rc != 0) {
            printf("LSynth failed\n");
        } else {
            printf("LSynth complete\n");
        }
    }
    
    /* close files */
    if (dat != stdin)
        fclose(dat);
    if (dst != stdout)
        fclose(dst);
        
    return rc;
}
